# -*- coding: utf-8 -*-
from threading import Thread
from time import sleep
from pathlib import Path

from natsort import natsorted
import psutil

from PyQt5 import QtGui
from PyQt5 import QtWidgets
from PyQt5 import QtCore

from ..ui import resource_rc as resource_rc
from ..ui.batch_widget import Ui_batch_widget

from ...mdf import MDF, SUPPORTED_VERSIONS
from ...signal import Signal
from ...blocks.utils import MdfException, extract_cncomment_xml
from ..utils import TERMINATED, run_thread_with_progress, setup_progress
from .search import SearchWidget
from .tree import TreeWidget
from .tree_item import TreeItem
from .list import MinimalListWidget


class BatchWidget(Ui_batch_widget, QtWidgets.QWidget):

    def __init__(self, ignore_value2text_conversions=False, *args, **kwargs):

        super().__init__(*args, **kwargs)
        self.setupUi(self)

        self.ignore_value2text_conversions = ignore_value2text_conversions

        self.progress = None
        self.files_list = MinimalListWidget()
        self.splitter.addWidget(self.files_list)

        self.raster_type_channel.toggled.connect(self.set_raster_type)

        for widget in (
              self.concatenate_format,
              self.convert_format,
              self.cut_format,
              self.resample_format,
              self.stack_format,
              self.extract_can_format,
        ):

            widget.insertItems(0, SUPPORTED_VERSIONS)

        for widget in (
              self.concatenate_split_size,
              self.convert_split_size,
              self.cut_split_size,
              self.resample_split_size,
              self.stack_split_size,
        ):

            widget.setValue(4)

        for widget in (
              self.concatenate_compression,
              self.convert_compression,
              self.cut_compression,
              self.resample_compression,
              self.stack_compression,
              self.extract_can_compression,
        ):

            widget.insertItems(
                0, ("no compression", "deflate", "transposed deflate")
            )

        self.concatenate_btn.clicked.connect(self.concatenate)
        self.convert_btn.clicked.connect(self.convert)
        self.cut_btn.clicked.connect(self.cut)
        self.resample_btn.clicked.connect(self.resample)
        self.scramble_btn.clicked.connect(self.scramble)
        self.stack_btn.clicked.connect(self.stack)
        self.extract_can_btn.clicked.connect(self.extract_can)
        self.extract_can_csv_btn.clicked.connect(self.extract_can_csv)

        self.load_can_database_btn.clicked.connect(self.load_can_database)

        self.empty_channels.insertItems(0, ("skip", "zeros"))
        self.empty_channels_can.insertItems(0, ("skip", "zeros"))
        self.mat_format.insertItems(0, ("4", "5", "7.3"))
        self.oned_as.insertItems(0, ("row", "column"))
        self.export_type.insertItems(0, ("csv", "hdf5", "mat", "parquet"))
        self.export_btn.clicked.connect(self.export)
        self.export_type.currentTextChanged.connect(self.export_changed)
        self.export_type.setCurrentIndex(-1)

        self.aspects.setCurrentIndex(0)
        self.setAcceptDrops(True)

    def set_raster_type(self, event):
        if self.raster_type_channel.isChecked():
            self.raster_channel.setEnabled(True)
            self.raster.setEnabled(False)
            self.raster.setValue(0)
        else:
            self.raster_channel.setEnabled(False)
            self.raster_channel.setCurrentIndex(0)
            self.raster.setEnabled(True)

    def export_changed(self, name):
        if name == 'parquet':
            self.export_compression.setEnabled(True)
            self.export_compression.clear()
            self.export_compression.addItems(['GZIP', 'SNAPPY'])
            self.export_compression.setCurrentIndex(-1)
        elif name == 'hdf5':
            self.export_compression.setEnabled(True)
            self.export_compression.clear()
            self.export_compression.addItems(["gzip", "lzf", "szip"])
            self.export_compression.setCurrentIndex(-1)
        elif name == 'mat':
            self.export_compression.setEnabled(True)
            self.export_compression.clear()
            self.export_compression.addItems(["enabled", "disabled"])
            self.export_compression.setCurrentIndex(-1)
        else:
            self.export_compression.clear()
            self.export_compression.setEnabled(False)

    def compute_cut_hints(self):
        # TODO : use master channel physical min and max values
        times = []
        groups_nr = len(self.mdf.groups)
        for i in range(groups_nr):
            master = self.mdf.get_master(i)
            if len(master):
                times.append(master[0])
                times.append(master[-1])
            QtWidgets.QApplication.processEvents()

        if len(times):
            time_range = min(times), max(times)

            self.cut_start.setRange(*time_range)
            self.cut_stop.setRange(*time_range)

            self.cut_interval.setText(
                "Cut interval ({:.6f}s - {:.6f}s)".format(*time_range)
            )
        else:
            self.cut_start.setRange(0, 0)
            self.cut_stop.setRange(0, 0)

            self.cut_interval.setText("Empty measurement")

    def update_progress(self, current_index, max_index):
        self.progress = current_index, max_index

    def close(self):
        mdf_name = self.mdf.name
        self.mdf.close()
        if self.file_name.suffix.lower() == ".dl3":
            mdf_name.unlink()

    def convert(self, event):
        version = self.convert_format.currentText()

        split = self.convert_split.checkState() == QtCore.Qt.Checked
        if split:
            split_size = int(self.convert_split_size.value() * 1024 * 1024)
        else:
            split_size = 0

        compression = self.convert_compression.currentIndex()

        count = self.files_list.count()

        delta = 100 / count

        progress = setup_progress(
            parent=self,
            title="Converting measurements",
            message=f'Converting "{count}" files to {version}',
            icon_name="convert",
        )

        files = self._prepare_files(progress)
        source_files = [
            Path(self.files_list.item(row).text())
            for row in range(count)
        ]

        for i, (file, source_file) in enumerate(zip(files, source_files)):

            progress.setLabelText(f'Converting file {i+1} of {count} to {version}')

            if not isinstance(file, MDF):

                # open file
                target = MDF
                kwargs = {
                    "name": file,
                    "callback": self.update_progress,
                }

                mdf = run_thread_with_progress(
                    self,
                    target=target,
                    kwargs=kwargs,
                    factor=0,
                    offset=int(i*delta),
                    progress=progress,
                )
            else:
                mdf = file

            target = mdf.convert
            kwargs = {"version": version}

            mdf = run_thread_with_progress(
                self,
                target=target,
                kwargs=kwargs,
                factor=int(delta/2),
                offset=int(i*delta),
                progress=progress,
            )

            if mdf is TERMINATED:
                progress.cancel()
                return

            mdf.configure(write_fragment_size=split_size)

            file_name = source_file.with_suffix(
                '.convert.mdf' if version < '4.00' else '.convert.mf4'
            )

            # then save it
            progress.setLabelText(f'Saving converted file {i+1} to "{file_name}"')

            target = mdf.save
            kwargs = {"dst": file_name, "compression": compression, "overwrite": True}

            run_thread_with_progress(
                self,
                target=target,
                kwargs=kwargs,
                factor=int(delta/2),
                offset=int((i+1/2)*delta),
                progress=progress,
            )

        self.progress = None
        progress.cancel()

    def resample(self, event):
        version = self.resample_format.currentText()

        if self.raster_type_channel.isChecked():
            raster = self.raster_channel.currentText()
        else:
            raster = self.raster.value()

        split = self.resample_split.checkState() == QtCore.Qt.Checked
        if split:
            split_size = int(self.resample_split_size.value() * 1024 * 1024)
        else:
            split_size = 0

        compression = self.resample_compression.currentIndex()
        time_from_zero = self.resample_time_from_zero.checkState() == QtCore.Qt.Checked

        count = self.files_list.count()

        delta = 100 / count

        progress = setup_progress(
            parent=self,
            title="Resampling measurements",
            message=f'Resampling "{count}" files',
            icon_name="convert",
        )

        files = self._prepare_files(progress)
        source_files = [
            Path(self.files_list.item(row).text())
            for row in range(count)
        ]

        for i, (file, source_file) in enumerate(zip(files, source_files)):

            if isinstance(raster, str):
                progress.setLabelText(f'Resampling file {i+1} of {count} using "{raster}" as raster')
            else:
                progress.setLabelText(f'Resampling file {i+1} of {count} to {raster:.3f}s')

            if not isinstance(file, MDF):

                # open file
                target = MDF
                kwargs = {
                    "name": file,
                    "callback": self.update_progress,
                }

                mdf = run_thread_with_progress(
                    self,
                    target=target,
                    kwargs=kwargs,
                    factor=0,
                    offset=int(i*delta),
                    progress=progress,
                )
            else:
                mdf = file

            target = mdf.resample
            kwargs = {
                "raster": raster,
                "version": version,
                "time_from_zero": time_from_zero,
            }

            mdf = run_thread_with_progress(
                self,
                target=target,
                kwargs=kwargs,
                factor=int(delta/2),
                offset=int(i*delta),
                progress=progress,
            )

            if mdf is TERMINATED:
                progress.cancel()
                return

            mdf.configure(write_fragment_size=split_size)

            file_name = source_file.with_suffix(
                '.resample.mdf' if version < '4.00' else '.resample.mf4'
            )

            # then save it
            progress.setLabelText(f'Saving resampled file {i+1} to "{file_name}"')

            target = mdf.save
            kwargs = {"dst": file_name, "compression": compression, "overwrite": True}

            run_thread_with_progress(
                self,
                target=target,
                kwargs=kwargs,
                factor=int(delta/2),
                offset=int((i+1/2)*delta),
                progress=progress,
            )

        self.progress = None
        progress.cancel()

    def cut(self, event):
        version = self.cut_format.currentText()
        start = self.cut_start.value()
        stop = self.cut_stop.value()
        time_from_zero = self.cut_time_from_zero.checkState() == QtCore.Qt.Checked

        if self.whence.checkState() == QtCore.Qt.Checked:
            whence = 1
        else:
            whence = 0

        split = self.cut_split.checkState() == QtCore.Qt.Checked
        if split:
            split_size = int(self.cut_split_size.value() * 1024 * 1024)
        else:
            split_size = 0

        compression = self.cut_compression.currentIndex()

        count = self.files_list.count()

        delta = 100 / count

        progress = setup_progress(
            parent=self,
            title="Cutting measurements",
            message=f'Cutting "{count}" files from {start}s to {stop}s',
            icon_name="cut",
        )

        files = self._prepare_files(progress)
        source_files = [
            Path(self.files_list.item(row).text())
            for row in range(count)
        ]

        for i, (file, source_file) in enumerate(zip(files, source_files)):

            progress.setLabelText(f'Cutting file {i+1} of {count} from {start}s to {stop}s')

            if not isinstance(file, MDF):

                # open file
                target = MDF
                kwargs = {
                    "name": file,
                    "callback": self.update_progress,
                }

                mdf = run_thread_with_progress(
                    self,
                    target=target,
                    kwargs=kwargs,
                    factor=0,
                    offset=int(i*delta),
                    progress=progress,
                )
            else:
                mdf = file

            # cut self.mdf
            target = mdf.cut
            kwargs = {
                "start": start,
                "stop": stop,
                "whence": whence,
                "version": version,
                "time_from_zero": time_from_zero,
            }

            mdf = run_thread_with_progress(
                self,
                target=target,
                kwargs=kwargs,
                factor=int(delta/2),
                offset=int(i*delta),
                progress=progress,
            )

            if mdf is TERMINATED:
                progress.cancel()
                return

            mdf.configure(write_fragment_size=split_size)

            file_name = source_file.with_suffix(
                '.cut.mdf' if version < '4.00' else '.cut.mf4'
            )

            # then save it
            progress.setLabelText(f'Saving cut file {i+1} to "{file_name}"')

            target = mdf.save
            kwargs = {"dst": file_name, "compression": compression, "overwrite": True}

            run_thread_with_progress(
                self,
                target=target,
                kwargs=kwargs,
                factor=int(delta/2),
                offset=int((i+1/2)*delta),
                progress=progress,
            )

        self.progress = None
        progress.cancel()

    def export(self, event):
        export_type = self.export_type.currentText()

        single_time_base = self.single_time_base.checkState() == QtCore.Qt.Checked
        time_from_zero = self.time_from_zero.checkState() == QtCore.Qt.Checked
        use_display_names = self.use_display_names.checkState() == QtCore.Qt.Checked
        empty_channels = self.empty_channels.currentText()
        mat_format = self.mat_format.currentText()
        raster = self.export_raster.value()
        oned_as = self.oned_as.currentText()
        reduce_memory_usage = self.reduce_memory_usage.checkState() == QtCore.Qt.Checked
        compression = self.export_compression.currentText()
        time_as_date = self.time_as_date.checkState() == QtCore.Qt.Checked

        count = self.files_list.count()

        delta = 100 / count

        progress = setup_progress(
            parent=self,
            title="Exporting measurements",
            message=f'Exporting "{count}" files to {export_type}',
            icon_name="export",
        )

        files = self._prepare_files(progress)
        source_files = [
            Path(self.files_list.item(row).text())
            for row in range(count)
        ]

        if export_type == 'csv':
            suffix = '.csv'
        elif export_type == "hdf5":
            suffix = '.hdf'
        elif export_type == "mat":
            suffix = '.mat'
        elif export_type == "parquet":
            suffix = '.parquet'

        for i, (file, source_file) in enumerate(zip(files, source_files)):

            progress.setLabelText(f'Exporting file {i+1} of {count} to {export_type}')

            if not isinstance(file, MDF):

                # open file
                target = MDF
                kwargs = {
                    "name": file,
                    "callback": self.update_progress,
                }

                mdf = run_thread_with_progress(
                    self,
                    target=target,
                    kwargs=kwargs,
                    factor=0,
                    offset=int(i*delta),
                    progress=progress,
                )
            else:
                mdf = file

            print(source_file)

            file_name = source_file.with_suffix(suffix)

            target = mdf.export
            kwargs={
                "fmt": export_type,
                "filename": file_name,
                "single_time_base": single_time_base,
                "use_display_names": use_display_names,
                "time_from_zero": time_from_zero,
                "empty_channels": empty_channels,
                "format": mat_format,
                "raster": raster,
                "oned_as": oned_as,
                "reduce_memory_usage": reduce_memory_usage,
                "compression": compression,
                "time_as_date": time_as_date,
                "ignore_value2text_conversions": self.ignore_value2text_conversions,
            }

            mdf = run_thread_with_progress(
                self,
                target=target,
                kwargs=kwargs,
                factor=int(delta/2),
                offset=int(i*delta),
                progress=progress,
            )

            if mdf is TERMINATED:
                progress.cancel()
                return

        self.progress = None
        progress.cancel()

    def filter(self, event):
        iterator = QtWidgets.QTreeWidgetItemIterator(self.filter_tree)

        group = -1
        index = 0
        channels = []
        while iterator.value():
            item = iterator.value()
            if item.parent() is None:
                iterator += 1
                group += 1
                index = 0
                continue

            if item.checkState(0) == QtCore.Qt.Checked:
                channels.append((None, group, index))

            index += 1
            iterator += 1

        version = self.filter_format.itemText(self.filter_format.currentIndex())

        if version < "4.00":
            filter = "MDF version 3 files (*.dat *.mdf)"
        else:
            filter = "MDF version 4 files (*.mf4)"

        split = self.filter_split.checkState() == QtCore.Qt.Checked
        if split:
            split_size = int(self.filter_split_size.value() * 1024 * 1024)
        else:
            split_size = 0

        self.mdf.configure(write_fragment_size=split_size)

        compression = self.filter_compression.currentIndex()

        file_name, _ = QtWidgets.QFileDialog.getSaveFileName(
            self, "Select output measurement file", "", filter
        )

        if file_name:
            progress = setup_progress(
                parent=self,
                title="Filtering measurement",
                message=f'Filtering selected channels from "{self.file_name}"',
                icon_name="filter",
            )

            # filtering self.mdf
            target = self.mdf.filter
            kwargs = {"channels": channels, "version": version}

            mdf = run_thread_with_progress(
                self,
                target=target,
                kwargs=kwargs,
                factor=66,
                offset=0,
                progress=progress,
            )

            if mdf is TERMINATED:
                progress.cancel()
                return

            mdf.configure(write_fragment_size=split_size)

            # then save it
            progress.setLabelText(f'Saving filtered file "{file_name}"')

            target = mdf.save
            kwargs = {"dst": file_name, "compression": compression, "overwrite": True}

            run_thread_with_progress(
                self,
                target=target,
                kwargs=kwargs,
                factor=34,
                offset=66,
                progress=progress,
            )

    def scramble(self, event):

        count = self.files_list.count()

        delta = 100 / count

        progress = setup_progress(
            parent=self,
            title="Scrambling measurements",
            message=f'Scrambling "{count}" files',
            icon_name="scramble",
        )

        files = self._prepare_files(progress)
        source_files = [
            Path(self.files_list.item(row).text())
            for row in range(count)
        ]

        for i, (file, source_file) in enumerate(zip(files, source_files)):

            progress.setLabelText(f'Scrambling file {i+1} of {count}')

            target = MDF.scramble
            kwargs = {
                "name": file,
                "callback": self.update_progress,
            }

            mdf = run_thread_with_progress(
                self,
                target=target,
                kwargs=kwargs,
                factor=0,
                offset=int(i*delta),
                progress=progress,
            )

            if mdf is TERMINATED:
                progress.cancel()
                return

        self.progress = None
        progress.cancel()

    def extract_can(self, event):
        version = self.extract_can_format.currentText()
        count = self.can_database_list.count()

        dbc_files = []
        for i in range(count):
            item = self.can_database_list.item(i)
            dbc_files.append(item.text())

        compression = self.extract_can_compression.currentIndex()
        ignore_invalid_signals = self.ignore_invalid_signals_mdf.checkState() == QtCore.Qt.Checked

        count = self.files_list.count()

        delta = 100 / count

        progress = setup_progress(
            parent=self,
            title="Extract CAN logging from measurements",
            message=f'Extracting CAN logging from "{count}" files',
            icon_name="down",
        )

        files = self._prepare_files(progress)
        source_files = [
            Path(self.files_list.item(row).text())
            for row in range(count)
        ]

        for i, (file, source_file) in enumerate(zip(files, source_files)):

            progress.setLabelText(f'Extracting CAN logging from file {i+1} of {count}')

            if not isinstance(file, MDF):

                # open file
                target = MDF
                kwargs = {
                    "name": file,
                    "callback": self.update_progress,
                }

                mdf = run_thread_with_progress(
                    self,
                    target=target,
                    kwargs=kwargs,
                    factor=0,
                    offset=int(i*delta),
                    progress=progress,
                )
            else:
                mdf = file

            target = mdf.extract_can_logging
            kwargs = {
                "dbc_files": dbc_files,
                "version": version,
                "ignore_invalid_signals": ignore_invalid_signals,
            }

            mdf = run_thread_with_progress(
                self,
                target=target,
                kwargs=kwargs,
                factor=int(delta/2),
                offset=int(i*delta),
                progress=progress,
            )

            if mdf is TERMINATED:
                progress.cancel()
                return

            file_name = source_file.with_suffix(
                '.can_logging.mdf' if version < '4.00' else '.can_logging.mf4'
            )

            # then save it
            progress.setLabelText(f'Saving extarcted CAN logging file {i+1} to "{file_name}"')

            target = mdf.save
            kwargs = {"dst": file_name, "compression": compression, "overwrite": True}

            run_thread_with_progress(
                self,
                target=target,
                kwargs=kwargs,
                factor=int(delta/2),
                offset=int((i+1/2)*delta),
                progress=progress,
            )

        self.progress = None
        progress.cancel()

    def extract_can_csv(self, event):
        version = self.extract_can_format.currentText()
        count = self.can_database_list.count()

        dbc_files = []
        for i in range(count):
            item = self.can_database_list.item(i)
            dbc_files.append(item.text())

        ignore_invalid_signals = self.ignore_invalid_signals_csv.checkState() == QtCore.Qt.Checked
        single_time_base = self.single_time_base_can.checkState() == QtCore.Qt.Checked
        time_from_zero = self.time_from_zero_can.checkState() == QtCore.Qt.Checked
        empty_channels = self.empty_channels_can.currentText()
        raster = self.export_raster.value()
        time_as_date = self.can_time_as_date.checkState() == QtCore.Qt.Checked

        count = self.files_list.count()

        delta = 100 / count

        progress = setup_progress(
            parent=self,
            title="Extract CAN logging from measurements to CSV",
            message=f'Extracting CAN logging from "{count}" files',
            icon_name="csv",
        )

        files = self._prepare_files(progress)
        source_files = [
            Path(self.files_list.item(row).text())
            for row in range(count)
        ]

        for i, (file, source_file) in enumerate(zip(files, source_files)):

            progress.setLabelText(f'Extracting CAN logging from file {i+1} of {count}')

            if not isinstance(file, MDF):

                # open file
                target = MDF
                kwargs = {
                    "name": file,
                    "callback": self.update_progress,
                }

                mdf = run_thread_with_progress(
                    self,
                    target=target,
                    kwargs=kwargs,
                    factor=0,
                    offset=int(i*delta),
                    progress=progress,
                )
            else:
                mdf = file

            target = mdf.extract_can_logging
            kwargs = {
                "dbc_files": dbc_files,
                "version": version,
                "ignore_invalid_signals": ignore_invalid_signals,
            }

            mdf = run_thread_with_progress(
                self,
                target=target,
                kwargs=kwargs,
                factor=int(delta/2),
                offset=int(i*delta),
                progress=progress,
            )

            if mdf is TERMINATED:
                progress.cancel()
                return

            file_name = source_file.with_suffix('.can_logging.csv')

            # then save it
            progress.setLabelText(f'Saving extarcted CAN logging file {i+1} to "{file_name}"')

            target = mdf.export
            kwargs = {
                "fmt": "csv",
                "filename": file_name,
                "single_time_base": single_time_base,
                "time_from_zero": time_from_zero,
                "empty_channels": empty_channels,
                "raster": raster,
                "time_as_date": time_as_date,
                "ignore_value2text_conversions": self.ignore_value2text_conversions,
            }

            run_thread_with_progress(
                self,
                target=target,
                kwargs=kwargs,
                factor=int(delta/2),
                offset=int((i+1/2)*delta),
                progress=progress,
            )

        self.progress = None
        progress.cancel()

    def load_can_database(self, event):
        file_names, _ = QtWidgets.QFileDialog.getOpenFileNames(
            self,
            "Select CAN database file",
            "",
            "ARXML or DBC (*.dbc *.axml)",
            "ARXML or DBC (*.dbc *.axml)",
        )

        for file_name in file_names:
            self.can_database_list.addItems(file_names)

    def concatenate(self, event):
        func = MDF.concatenate
        operation = "Concatenating"

        version = self.concatenate_format.currentText()

        sync = self.concatenate_sync.checkState() == QtCore.Qt.Checked
        add_samples_origin = self.concatenate_add_samples_origin.checkState() == QtCore.Qt.Checked

        if version < "4.00":
            filter = "MDF version 3 files (*.dat *.mdf)"
            suffix = '.mdf'
        else:
            filter = "MDF version 4 files (*.mf4)"
            suffix = '.mf4'

        split = self.concatenate_split.checkState() == QtCore.Qt.Checked
        if split:
            split_size = int(self.concatenate_split_size.value() * 1024 * 1024)
        else:
            split_size = 0

        compression = self.concatenate_compression.currentIndex()

        output_file_name, _ = QtWidgets.QFileDialog.getSaveFileName(
            self, "Select output measurement file", "", f'{filter};;All files (*.*)', filter
        )

        if output_file_name:
            output_file_name = Path(output_file_name).with_suffix(suffix)

            progress = setup_progress(
                parent=self,
                title=f"{operation} measurements",
                message=f"{operation} files and saving to {version} format",
                icon_name="stack",
            )

            files = self._prepare_files(progress)

            target = func
            kwargs = {
                "files": files,
                "outversion": version,
                "callback": self.update_progress,
                "sync": sync,
                "add_samples_origin": add_samples_origin,
            }

            mdf = run_thread_with_progress(
                self,
                target=target,
                kwargs=kwargs,
                factor=50,
                offset=0,
                progress=progress,
            )

            if mdf is TERMINATED:
                progress.cancel()
                return

            mdf.configure(write_fragment_size=split_size)

            # save it
            progress.setLabelText(f'Saving output file "{output_file_name}"')

            target = mdf.save
            kwargs = {"dst": output_file_name, "compression": compression, "overwrite": True}

            run_thread_with_progress(
                self,
                target=target,
                kwargs=kwargs,
                factor=50,
                offset=50,
                progress=progress,
            )

            progress.cancel()

    def stack(self, event):
        func = MDF.stack
        operation = "Stacking"

        version = self.stack_format.currentText()

        sync = self.stack_sync.checkState() == QtCore.Qt.Checked
        add_samples_origin = self.stack_add_samples_origin.checkState() == QtCore.Qt.Checked

        if version < "4.00":
            filter = "MDF version 3 files (*.dat *.mdf)"
            suffix = '.mdf'
        else:
            filter = "MDF version 4 files (*.mf4)"
            suffix = '.mf4'

        split = self.stack_split.checkState() == QtCore.Qt.Checked
        if split:
            split_size = int(self.stack_split_size.value() * 1024 * 1024)
        else:
            split_size = 0

        compression = self.stack_compression.currentIndex()

        output_file_name, _ = QtWidgets.QFileDialog.getSaveFileName(
            self, "Select output measurement file", "", f'{filter};;All files (*.*)', filter
        )

        if output_file_name:
            output_file_name = Path(output_file_name).with_suffix(suffix)

            progress = setup_progress(
                parent=self,
                title=f"{operation} measurements",
                message=f"{operation} files and saving to {version} format",
                icon_name="stack",
            )

            files = self._prepare_files(progress)

            target = func
            kwargs = {
                "files": files,
                "outversion": version,
                "callback": self.update_progress,
                "sync": sync,
                "add_samples_origin": add_samples_origin,
            }

            mdf = run_thread_with_progress(
                self,
                target=target,
                kwargs=kwargs,
                factor=50,
                offset=0,
                progress=progress,
            )

            if mdf is TERMINATED:
                progress.cancel()
                return

            mdf.configure(write_fragment_size=split_size)

            # save it
            progress.setLabelText(f'Saving output file "{output_file_name}"')

            target = mdf.save
            kwargs = {"dst": output_file_name, "compression": compression, "overwrite": True}

            run_thread_with_progress(
                self,
                target=target,
                kwargs=kwargs,
                factor=50,
                offset=50,
                progress=progress,
            )

            progress.cancel()

    def _prepare_files(self, progress):
        count = self.files_list.count()

        files = [Path(self.files_list.item(row).text()) for row in range(count)]

        for i, file_name in enumerate(files):
            if file_name.suffix.lower() == ".erg":
                progress.setLabelText(f"Converting file {i+1} of {count} from erg to mdf")
                try:
                    from mfile import ERG

                    files[i] = ERG(file_name).export_mdf()
                except Exception as err:
                    print(err)
                    return
            else:

                if file_name.suffix.lower() == ".dl3":
                    progress.setLabelText(f"Converting file {i+1} of {count} from dl3 to mdf")
                    datalyser_active = any(
                        proc.name() == 'Datalyser3.exe'
                        for proc in psutil.process_iter()
                    )
                    try:
                        import win32com.client

                        index = 0
                        while True:
                            mdf_name = file_name.with_suffix(f".{index}.mdf")
                            if mdf_name.exists():
                                index += 1
                            else:
                                break

                        datalyser = win32com.client.Dispatch("Datalyser3.Datalyser3_COM")
                        if not datalyser_active:
                            try:
                                datalyser.DCOM_set_datalyser_visibility(False)
                            except:
                                pass
                        datalyser.DCOM_convert_file_mdf_dl3(file_name, str(mdf_name), 0)
                        if not datalyser_active:
                            datalyser.DCOM_TerminateDAS()
                        files[i] = mdf_name
                    except Exception as err:
                        print(err)
                        return

        return files



# -*- coding: utf-8 -*-
import logging

try:
    from PyQt5 import QtWidgets
    from .widgets.plot_standalone import PlotWindow
    QT = True
except ImportError:
    QT = False


logger = logging.getLogger("asammdf")


def plot(signals, title=''):
    """ create a stand-alone plot using the input signal or signals

    Arguments
    ---------
    signals : iterable | Signal

    title (""): str
        window title

    """

    if QT:
        app = QtWidgets.QApplication([])
        app.setOrganizationName("py-asammdf")
        app.setOrganizationDomain("py-asammdf")
        app.setApplicationName("py-asammdf")
        main = PlotWindow(signals)
        if title.strip():
            main.setWindowTitle(title.strip())
        else:
            if isinstance(signals, (tuple, list)):
                main.setWindowTitle(', '.join(sig.name for sig in signals))
            else:
                main.setWindowTitle(signals.name)

        app.setStyle(QtWidgets.QStyleFactory.create("Fusion"))

        app.exec_()

    else:
        logging.warning("Signal plotting requires pyqtgraph or matplotlib")
        raise Exception("Signal plotting requires pyqtgraph or matplotlib")



import os
import struct

from enum import IntFlag
from typing import Optional, Tuple, BinaryIO, Union, Dict

from asammdf.blocks.utils import MdfException


class FinalizationFlags(IntFlag):
    CG_CA_CYCLE_COUNTERS = 1 << 0
    SR_CYCLE_COUNTERS = 1 << 1
    DT_LENGTH = 1 << 2
    RD_LENGTH = 1 << 3
    UPDATE_DL = 1 << 4
    VLSD_CG_DATA_BYTES = 1 << 5
    VLSD_CG_OFFSET = 1 << 6


class MdfHeader:
    """Representation of the MDF header.
    """
    _fmt = "<4s4xQQ"
    _fmt_length = 24

    def __init__(self, identification_str: str = None, length: int = 0, link_count: int = 0):
        self.identification_str = identification_str
        self.length = length
        self.link_count = link_count

    def __bytes__(self):
        return struct.pack(self._fmt, self.identification_str, self.length, self.link_count)

    def __len__(self):
        return self._fmt_length

    pass


class MdfBlock():
    """Generic representation of a MDF4 block (Except the ID block).
    """
    def __init__(self, header: MdfHeader):
        self.header: MdfHeader = header
        self.file_location: int = 0
        self.links: Tuple[int, ...] = (0,) * header.link_count

        return

    def load(self, stream: BinaryIO):
        """Function responsible for loading anything past the header of the block.

        Expects the stream position to be just after the header.

        Parameters
        ----------
        stream : BinaryIO
            Access to block data
        """
        # Load the addresses of the block links.
        self.links = struct.unpack(f"<{self.header.link_count}Q", stream.read(8 * self.header.link_count))

        return

    def __bytes__(self):
        # Aggregate the result of the header and the links.
        result = bytes(self.header)
        result += struct.pack(f"<{self.header.link_count}Q", *self.links)

        return result

    pass


class CG_Block(MdfBlock):
    _fmt = "<QQHH4xLL"
    _fmt_length = 32

    def __init__(self, header: MdfHeader = None):
        if not header:
            header = MdfHeader(b"##CG", 104, 6)
        super().__init__(header)

        # Data fields.
        self.record_id = 0
        self.cycle_count = 0
        self.flags = 0
        self.path_separator = 0
        self.data_bytes = 0
        self.inval_bytes = 0

        return

    @property
    def is_vlsd(self) -> bool:
        """Is this a VLSD block.

        Returns
        -------
        bool
            True if VLSD block, False otherwise.
        """
        return (self.flags & (1 << 0)) == 1

    @property
    def get_next_cg_block_address(self) -> int:
        """Get the next CG block address in this chain of CG blocks.

        Returns
        -------
        int
            The address of the next CG block, or 0 if none exists.
        """
        return self.links[0]

    @property
    def get_first_cn_block_address(self) -> int:
        """Get the first CN block address for this CG block.

        Returns
        -------
        int
            The address of the first CN block, or 0 if none exists.
        """
        return self.links[1]

    def load(self, stream: BinaryIO):
        # Load links via super class.
        super().load(stream)

        # Load data.
        (
            self.record_id,
            self.cycle_count,
            self.flags,
            self.path_separator,
            self.data_bytes,
            self.inval_bytes,
        ) = struct.unpack(self._fmt, stream.read(self._fmt_length))

        return

    def __bytes__(self):
        # Aggregate the result of the header and links from the super class, and the data from this class.
        result = super().__bytes__()

        result += struct.pack(
            self._fmt,
            self.record_id,
            self.cycle_count,
            self.flags,
            self.path_separator,
            self.data_bytes,
            self.inval_bytes,
        )

        return result

    pass


class CN_Block(MdfBlock):
    _fmt = "<4B4LBxH6d"
    _fmt_length = 72

    def __init__(self, header: MdfHeader = None):
        if not header:
            header = MdfHeader(b"##CN", 8, 160)
        super().__init__(header)

        # Data fields.
        self.channel_type = 0
        self.sync_type = 0
        self.data_type = 0
        self.bit_offset = 0
        self.byte_offset = 0
        self.bit_count = 0
        self.flags = 0
        self.pos_invalidation_bit = 0
        self.precision = 0
        self.attachment_nr = 0
        self.min_raw_value = 0
        self.max_raw_value = 0
        self.lower_limit = 0
        self.upper_limit = 0
        self.lower_ext_limit = 0
        self.upper_ext_limit = 0

        return

    @property
    def get_next_cn_block_address(self):
        """Get the next CN block address in this chain of CN blocks.

        Returns
        -------
        int
            The address of the next CN block, or 0 if none exists.
        """
        return self.links[0]

    def load(self, stream: BinaryIO):
        # Load links via super class.
        super().load(stream)

        # Load data.
        (
            self.channel_type,
            self.sync_type,
            self.data_type,
            self.bit_offset,
            self.byte_offset,
            self.bit_count,
            self.flags,
            self.pos_invalidation_bit,
            self.precision,
            self.attachment_nr,
            self.min_raw_value,
            self.max_raw_value,
            self.lower_limit,
            self.upper_limit,
            self.lower_ext_limit,
            self.upper_ext_limit,
        ) = struct.unpack(self._fmt, stream.read(self._fmt_length))

        return

    def __bytes__(self):
        # Aggregate the result of the header and links from the super class, and the data from this class.
        result = super().__bytes__()

        result += struct.pack(
            self._fmt,
            self.channel_type,
            self.sync_type,
            self.data_type,
            self.bit_offset,
            self.byte_offset,
            self.bit_count,
            self.flags,
            self.pos_invalidation_bit,
            self.precision,
            self.attachment_nr,
            self.min_raw_value,
            self.max_raw_value,
            self.lower_limit,
            self.upper_limit,
            self.lower_ext_limit,
            self.upper_ext_limit,
        )

        return result

    pass


class DG_Block(MdfBlock):
    _fmt = "<B7x"
    _fmt_length = 8

    def __init__(self, header: MdfHeader = None):
        if not header:
            header = MdfHeader(b"##DG", 64, 4)
        super().__init__(header)

        # Data fields.
        self.record_size = 0

        return

    @property
    def get_next_dg_block_address(self):
        """Get the next DG block address in this chain of DG blocks.

        Returns
        -------
        int
            The address of the next DG block, or 0 if none exists.
        """
        return self.links[0]

    @property
    def get_cg_block_address(self):
        """Get the address of the linked CG block for this DG block.

        Returns
        -------
        int
            The address of the linked CG block, or 0 if none exists.
        """
        return self.links[1]

    @property
    def get_dt_block_address(self):
        """Get the address of the linked DT block for this DG block.

        Returns
        -------
        int
            The address of the linked DT block, or 0 if none exists.
        """
        return self.links[2]

    def load(self, stream: BinaryIO):
        # Load links via super class.
        super().load(stream)

        # Load data.
        self.record_size, = struct.unpack(self._fmt, stream.read(self._fmt_length))
        pass

    def __bytes__(self):
        # Aggregate the result of the header and links from the super class, and the data from this class.
        result = super().__bytes__()

        result += struct.pack(
            self._fmt,
            self.record_size,
        )

        return result

    pass


class HD_Block(MdfBlock):
    _fmt = "<Q2h3Bx2d"
    _fmt_length = 32

    def __init__(self, header: MdfHeader = None):
        if not header:
            header = MdfHeader(b"##HD", 104, 6)
        super().__init__(header)

        # Data fields.
        self.abs_time = 0
        self.tz_offset = 0
        self.daylight_save_time = 0
        self.time_flags = 0
        self.time_quality = 0
        self.flags = 0
        self.start_angle = 0
        self.start_distance = 0

        return

    @property
    def get_dg_block_address(self):
        """Get the address of the linked DG block for this HD block.

        Returns
        -------
        int
            The address of the linked DG block, or 0 if none exists.
        """
        return self.links[1]

    def load(self, stream: BinaryIO):
        # Load links via super class.
        super().load(stream)

        # Load data.
        (
            self.abs_time,
            self.tz_offset,
            self.daylight_save_time,
            self.time_flags,
            self.time_quality,
            self.flags,
            self.start_angle,
            self.start_distance,
        ) = struct.unpack(self._fmt, stream.read(self._fmt_length))
        pass

    def __bytes__(self):
        # Aggregate the result of the header and links from the super class, and the data from this class.
        result = super().__bytes__()

        result += struct.pack(
            self._fmt,
            self.abs_time,
            self.tz_offset,
            self.daylight_save_time,
            self.time_flags,
            self.time_quality,
            self.flags,
            self.start_angle,
            self.start_distance,
        )

        return result

    pass


class FinalizationShim:
    """Shim class to insert between un-finalized MDF files and asammdf.
    """
    def __init__(self, parent: BinaryIO, finalization_flags: int):
        super().__init__()

        # Check that the requested flags are supported.
        supported_flags = FinalizationFlags.VLSD_CG_DATA_BYTES
        supported_flags |= FinalizationFlags.CG_CA_CYCLE_COUNTERS
        supported_flags |= FinalizationFlags.DT_LENGTH

        if finalization_flags & ~supported_flags:
            raise MdfException("Unsupported finalization options detected")

        self._parent = parent
        self._blocks: {int, MdfBlock} = {}
        self._flags = finalization_flags

        # Determine the file size by seeking to the end of the stream.
        self._parent.seek(0, os.SEEK_END)
        self._file_size = self._parent.tell()

        # Keep track of the local/shimmed position.
        self._location = 0

        return

    def load_blocks(self):
        """Creates a tree of all the blocks.
        """

        # Load all blocks recursively, starting with the HD block at address 64.
        self._load_block(64)

    def finalize(self):
        """Attempts to perform in-memory finalization.
        """
        # Look at the finalization flags.
        if self._flags & FinalizationFlags.DT_LENGTH:
            self._handle_dt()
            self._flags = ~FinalizationFlags.DT_LENGTH & self._flags
            pass
        if (self._flags & FinalizationFlags.CG_CA_CYCLE_COUNTERS) or (self._flags & FinalizationFlags.VLSD_CG_DATA_BYTES):
            # Since all data records have to be read to get valid cycle counts, data bytes can be updated for "free".
            self._handle_cg_counters()
            self._flags = ~FinalizationFlags.CG_CA_CYCLE_COUNTERS & self._flags
            self._flags = ~FinalizationFlags.VLSD_CG_DATA_BYTES & self._flags
            pass

        return

    # Overload methods of BinaryIO.
    def close(self):
        # Delegate to enclosed stream.
        self._parent.close()
        return

    def read(self, size: Optional[int] = ...) -> bytes:
        if size == ...:
            size = 0

        # Does the shim contain a block in the requested memory area?
        match: Optional[MdfBlock] = None

        for block in self._blocks.values():
            start = block.file_location
            end = start + block.header.length

            if start <= self._location < end:
                match = block
                break

        result = None

        if match is not None:
            # Replace returned data with data from the loaded block in the shim.
            data = bytes(match)
            offset = self._location - match.file_location

            if len(data) - offset > size:
                result = data[offset:offset+size]
                self._location += size
                size = 0
            else:
                result = data[offset:]
                self._location += len(result)
                size -= len(result)
            pass

        # Supply with additional data from the original file if necessary.
        self._parent.seek(self._location)
        additional_data = self._parent.read(size)

        result += additional_data

        return result

    def seek(self, offset: int, whence: int = ...) -> int:
        if whence == ...:
            whence = 0

        if whence == 0:
            self._location = offset
        elif whence == 2:
            self._location = self._file_size - offset
        return self._parent.seek(offset, whence)

    # Python overrides.
    def __iter__(self):
        return self

    def __next__(self):
        self._parent.seek(self._location)
        result = self._parent.read(1)
        self._location += 1

        return result

    def _handle_dt(self):
        """Update the length of the last DT block. Must be bounded by end of file (EOF) or another block.
        """

        dt_block: Optional[MdfBlock] = None

        # Start by iterating over the loaded blocks in reverse, looking for a DT block.
        for key in reversed(sorted(self._blocks.keys())):
            block = self._blocks[key]

            if block.header.identification_str == b"##DT":
                dt_block = block
                break

        if dt_block is None:
            raise RuntimeError("Requested to finalize DT block, but no DT blocks found in the file.")

        # Set the initial guess at a bound to the EOF.
        bound = self._file_size

        # Determine if bound by another block.
        keys = sorted(self._blocks.keys())

        for i, key in enumerate(keys):
            if key == dt_block.file_location:
                # Check if a block exists after this.
                if len(self._blocks) - 1 != i:
                    bound = keys[i + 1]
                    break

        # Update block size.
        dt_block.header.length = bound - dt_block.file_location

        return

    def _handle_cg_counters(self):
        """Update CG counters, set data bytes for VLSD CG blocks.
        """

        # Find all DG blocks and their corresponding data block. Loop over all CG blocks in each DG block.
        hd_block = self._blocks[64]
        dg_block: DG_Block = self._blocks.get(hd_block.links[0])

        while dg_block is not None:
            # Find all CG blocks for this DG block.
            cg_blocks: [CG_Block] = []
            cg_block: CG_Block = self._blocks.get(dg_block.get_cg_block_address)

            while cg_block is not None:
                cg_blocks.append(cg_block)
                cg_block = self._blocks.get(cg_block.get_next_cg_block_address)
                pass

            # Determine the number of cycles in each, as well as number of bytes in CG VLSD blocks. This is done by
            # looping over the records in the DT block, so the record size as well as record IDs along with the sizes
            # of the static data can to be determined.
            dg_record_size = dg_block.record_size
            dg_record_fmt = ""
            if dg_record_size == 1:
                dg_record_fmt = "<B"
            elif dg_record_size == 2:
                dg_record_fmt = "<H"
            elif dg_record_size == 4:
                dg_record_fmt = "<L"
            elif dg_record_size == 8:
                dg_record_fmt = "<Q"

            size_map = {}

            for cg_block in cg_blocks:
                cg_block_id = cg_block.record_id

                if cg_block.is_vlsd:
                    # Invalid size to signify this is a VLSD record, which should obtain the size using another method.
                    cg_block_size = -1
                else:
                    # Count the number of bits used by all sub-channels. This is done by taking the highest bit/byte
                    # offset, and utilising the bit count there.
                    cn_block: CN_Block = self._blocks.get(cg_block.get_first_cn_block_address)
                    highest_bit_value = 0

                    while cn_block is not None:
                        bit_value = cn_block.byte_offset * 8 + cn_block.bit_offset + cn_block.bit_count

                        if bit_value > highest_bit_value:
                            highest_bit_value = bit_value

                        cn_block = self._blocks.get(cn_block.get_next_cn_block_address)
                        pass

                    # Convert from bits to a byte value.
                    cg_block_size = highest_bit_value // 8

                    remainder = highest_bit_value % 8
                    if remainder != 0:
                        cg_block_size += 1
                    pass

                # Store the result as a new record in the size dictionary.
                size_map[cg_block_id] = cg_block_size

            # Loop over the records in the DT block, and extract record counts and data lengths.
            dt_location = 0
            dt_block = self._blocks.get(dg_block.get_dt_block_address)
            dt_block_size = dt_block.header.length - len(dt_block.header)

            self._parent.seek(dt_block.file_location + len(dt_block.header))

            # Prepare storage for cycle counters and data sizes.
            cycle_counters: Dict[int, int] = {}
            data_bytes: Dict[int, int] = {}
            for key in size_map.keys():
                cycle_counters[key] = 0
                data_bytes[key] = 0
                pass

            while dt_location < dt_block_size:
                # Read record id.
                record_id, = struct.unpack(dg_record_fmt, self._parent.read(dg_record_size))
                dt_location += dg_record_size

                # Increment the corresponding cycle counter.
                cycle_counters[record_id] += 1

                # Determine if this is static or dynamic data.
                record_length = size_map.get(record_id)

                if record_length < 0:
                    # VLSD data, read 4 bytes of length information from the data record.
                    record_length, = struct.unpack("<L", self._parent.read(4))
                    dt_location += 4
                    pass

                # Skip over the data, but store the size.
                self._parent.read(record_length)
                data_bytes[record_id] += record_length
                dt_location += record_length
                pass

            # Update the cycle counters and data bytes if requested.
            for cg_block in cg_blocks:
                record_id = cg_block.record_id

                if self._flags & FinalizationFlags.CG_CA_CYCLE_COUNTERS:
                    cg_block.cycle_count = cycle_counters.get(record_id)

                if cg_block.is_vlsd and (self._flags & FinalizationFlags.VLSD_CG_DATA_BYTES):
                    cg_block.data_bytes = (data_bytes.get(record_id) & 0x00000000FFFFFFFF)
                    cg_block.inval_bytes = (data_bytes.get(record_id) & 0xFFFFFFFF00000000) << 32

            dg_block = self._blocks.get(dg_block.get_next_dg_block_address)
            pass

        return

    def _load_block(self, address: int):
        """Load a MDF block and all linked blocks.

        Parameters
        ----------
        address : int
            The file address to load the block at.

        Returns
        -------
        MDfBlock
            The loaded block.
        """
        # Read the header bytes.
        self._parent.seek(address)

        data = struct.unpack("<4s4x2Q", self._parent.read(24))
        header = MdfHeader(*data)

        # Determine the block type.
        block = None
        if header.identification_str == b"##CG":
            block = CG_Block(header)
        elif header.identification_str == b"##CN":
            block = CN_Block(header)
        elif header.identification_str == b"##DG":
            block = DG_Block(header)
        elif header.identification_str == b"##HD":
            block = HD_Block(header)
        else:
            block = MdfBlock(header)

        block.file_location = address
        block.load(self._parent)

        self._blocks[address] = block

        for link_address in block.links:
            if link_address == 0:
                continue

            self._load_block(link_address)
            pass

        return block

    pass
